// Populated by CMake, and installed to the host system
#ifndef NOTCURSES_VERSION_HEADER
#define NOTCURSES_VERSION_HEADER

#define NOTCURSES_VERNUM_MAJOR 3
#define NOTCURSES_VERNUM_MINOR 0
#define NOTCURSES_VERNUM_PATCH 17
#define NOTCURSES_VERNUM_TWEAK 

#define NOTCURSES_VERSION_COMPARABLE(major, minor, patch) \
  (((major) << 16u) + ((minor) << 8u) + (patch))

#define NOTCURSES_VERNUM_ORDERED NOTCURSES_VERSION_COMPARABLE( \
  NOTCURSES_VERNUM_MAJOR, NOTCURSES_VERNUM_MINOR, NOTCURSES_VERNUM_PATCH)

#define NOTCURSES_VERSION_MAJOR "3"
#define NOTCURSES_VERSION_MINOR "0"
#define NOTCURSES_VERSION_PATCH "17"
#define NOTCURSES_VERSION_TWEAK ""

#endif
